@@ -130,8 +130,8 @@ describe "Sequel Mock Adapter" do
   end
 
   specify "should raise Error if given an invalid object to fetch" do
-    proc{Sequel.mock(:fetch=>Class.new).get(1)}.should raise_error(Sequel::Error)
-    proc{Sequel.mock(:fetch=>Object.new).get(1)}.should raise_error(Sequel::Error)
+    proc{Sequel.mock(:fetch=>Class.new).get(:a)}.should raise_error(Sequel::Error)
+    proc{Sequel.mock(:fetch=>Object.new).get(:a)}.should raise_error(Sequel::Error)
   end
 
   specify "should be able to set the number of rows modified by update and delete using :numrows option as an integer" do
@@ -417,12 +417,12 @@ describe "Sequel Mock Adapter" do
 
   specify "should be able to load dialects based on the database name" do
     begin
-      qi = class Sequel::Database; @@quote_identifiers; end
-      ii = class Sequel::Database; @@identifier_input_method; end
-      io = class Sequel::Database; @@identifier_output_method; end
+      qi = class Sequel::Database; @quote_identifiers; end
+      ii = class Sequel::Database; @identifier_input_method; end
+      io = class Sequel::Database; @identifier_output_method; end
       Sequel.quote_identifiers = nil
-      class Sequel::Database; @@identifier_input_method=nil; end
-      class Sequel::Database; @@identifier_output_method=nil; end
+      class Sequel::Database; @identifier_input_method=nil; end
+      class Sequel::Database; @identifier_output_method=nil; end
       Sequel.mock(:host=>'access').select(Date.new(2011, 12, 13)).sql.should == 'SELECT #2011-12-13#'
       Sequel.mock(:host=>'db2').select(1).sql.should == 'SELECT 1 FROM "SYSIBM"."SYSDUMMY1"'
       Sequel.mock(:host=>'firebird')[:a].distinct.limit(1, 2).sql.should == 'SELECT DISTINCT FIRST 1 SKIP 2 * FROM "A"'
@@ -434,16 +434,25 @@ describe "Sequel Mock Adapter" do
       Sequel.mock(:host=>'sqlite')[:a___b].sql.should == "SELECT * FROM `a` AS 'b'"
     ensure
       Sequel.quote_identifiers = qi
-      Sequel::Database.send(:class_variable_set, :@@identifier_input_method, ii)
-      Sequel::Database.send(:class_variable_set, :@@identifier_output_method, io)
+      Sequel::Database.send(:instance_variable_set, :@identifier_input_method, ii)
+      Sequel::Database.send(:instance_variable_set, :@identifier_output_method, io)
     end
   end
 
-  specify "should automatically set version for postgres" do
+  specify "should automatically set version for postgres and mssql" do
     Sequel.mock(:host=>'postgres').server_version.should == 90103
+    Sequel.mock(:host=>'mssql').server_version.should == 10000000
   end
 
   specify "should stub out the primary_key method for postgres" do
     Sequel.mock(:host=>'postgres').primary_key(:t).should == :id
   end
+
+  specify "should stub out the bound_variable_arg method for postgres" do
+    Sequel.mock(:host=>'postgres').bound_variable_arg(:t, nil).should == :t
+  end
+
+  specify "should handle creating tables on oracle" do
+    proc{Sequel.mock(:host=>'oracle').create_table(:a){String :b}}.should_not raise_error
+  end
 end
