@@ -113,6 +113,22 @@ describe "DB#create_table" do
     @db.sqls.should == ['CREATE TABLE cats (id serial PRIMARY KEY)']
   end
 
+  specify "should allow naming primary key constraint with :primary_key_constraint_name option" do
+    @db.create_table(:cats) do
+      primary_key :id, :primary_key_constraint_name=>:foo
+    end
+    @db.sqls.should == ['CREATE TABLE cats (id integer CONSTRAINT foo PRIMARY KEY AUTOINCREMENT)']
+  end
+
+  specify "should handling splitting named column constraints into table constraints if unsupported" do
+    def @db.supports_named_column_constraints?; false end
+    @db.create_table(:cats) do
+      primary_key :id, :primary_key_constraint_name=>:foo
+      foreign_key :cat_id, :cats, :unique=>true, :unique_constraint_name=>:bar, :foreign_key_constraint_name=>:baz, :deferrable=>true, :key=>:foo_id, :on_delete=>:cascade, :on_update=>:restrict
+    end
+    @db.sqls.should == ['CREATE TABLE cats (id integer AUTOINCREMENT, cat_id integer, CONSTRAINT foo PRIMARY KEY (id), CONSTRAINT baz FOREIGN KEY (cat_id) REFERENCES cats(foo_id) ON DELETE CASCADE ON UPDATE RESTRICT DEFERRABLE INITIALLY DEFERRED, CONSTRAINT bar UNIQUE (cat_id))']
+  end
+
   specify "should accept and literalize default values" do
     @db.create_table(:cats) do
       integer :id, :default => 123
@@ -147,6 +163,13 @@ describe "DB#create_table" do
     @db.sqls.should == ["CREATE TABLE cats (id integer, name text UNIQUE)"]
   end
   
+  specify "should allow naming unique constraint with :unique_constraint_name option" do
+    @db.create_table(:cats) do
+      text :name, :unique => true, :unique_constraint_name=>:foo
+    end
+    @db.sqls.should == ["CREATE TABLE cats (name text CONSTRAINT foo UNIQUE)"]
+  end
+  
   specify "should handle not deferred unique constraints" do
     @db.create_table(:cats) do
       integer :id
@@ -227,6 +250,13 @@ describe "DB#create_table" do
     @db.sqls.should == ["CREATE TABLE cats (project_id integer DEFAULT 3 REFERENCES projects)"]
   end
   
+  specify "should allowing naming foreign key constraint with :foreign_key_constraint_name option" do
+    @db.create_table(:cats) do
+      foreign_key :project_id, :projects, :foreign_key_constraint_name=>:foo
+    end
+    @db.sqls.should == ["CREATE TABLE cats (project_id integer CONSTRAINT foo REFERENCES projects)"]
+  end
+  
   specify "should raise an error if the table argument to foreign_key isn't a hash, symbol, or nil" do
     proc{@db.create_table(:cats){foreign_key :project_id, Object.new, :default=>3}}.should raise_error(Sequel::Error)
   end
@@ -359,7 +389,7 @@ describe "DB#create_table" do
     @db.sqls.should == ["CREATE TABLE cats (id integer)", "CREATE INDEX cats_id_index ON cats (id)"]
   end
 
-  specify "should accept unique index definitions" do
+  specify "should accept unique constraint definitions" do
     @db.create_table(:cats) do
       text :name
       unique :name
@@ -413,7 +443,7 @@ describe "DB#create_table" do
   end
 
   specify "should ignore errors if the database raises an error on an index creation statement and the :ignore_index_errors option is used" do
-    @db.meta_def(:execute_ddl){|*a| raise Sequel::DatabaseError if /blah/.match(a.first); super(*a)}
+    meta_def(@db, :execute_ddl){|*a| raise Sequel::DatabaseError if /blah/.match(a.first); super(*a)}
     lambda{@db.create_table(:cats){Integer :id; index :blah; index :id}}.should raise_error(Sequel::DatabaseError)
     @db.sqls.should == ['CREATE TABLE cats (id integer)']
     lambda{@db.create_table(:cats, :ignore_index_errors=>true){Integer :id; index :blah; index :id}}.should_not raise_error(Sequel::DatabaseError)
@@ -472,11 +502,19 @@ describe "DB#create_table" do
   specify "should accept unnamed constraint definitions with blocks" do
     @db.create_table(:cats) do
       integer :score
-      check{(x.sql_number > 0) & (y.sql_number < 1)}
+      check{(x > 0) & (y < 1)}
     end
     @db.sqls.should == ["CREATE TABLE cats (score integer, CHECK ((x > 0) AND (y < 1)))"]
   end
 
+  specify "should accept unnamed constraint definitions with function calls" do
+    @db.create_table(:cats) do
+      integer :score
+      check{f(x)}
+    end
+    @db.sqls.should == ["CREATE TABLE cats (score integer, CHECK (f(x)))"]
+  end
+
   specify "should accept unnamed constraint definitions" do
     @db.create_table(:cats) do
       check 'price < ?', 100
@@ -484,6 +522,13 @@ describe "DB#create_table" do
     @db.sqls.should == ["CREATE TABLE cats (CHECK (price < 100))"]
   end
 
+  specify "should accept arrays of pairs constraints" do
+    @db.create_table(:cats) do
+      check [[:price, 100]]
+    end
+    @db.sqls.should == ["CREATE TABLE cats (CHECK (price = 100))"]
+  end
+
   specify "should accept hash constraints" do
     @db.create_table(:cats) do
       check :price=>100
@@ -491,6 +536,13 @@ describe "DB#create_table" do
     @db.sqls.should == ["CREATE TABLE cats (CHECK (price = 100))"]
   end
 
+  specify "should accept array constraints" do
+    @db.create_table(:cats) do
+      check [Sequel.expr(:x) > 0, Sequel.expr(:y) < 1]
+    end
+    @db.sqls.should == ["CREATE TABLE cats (CHECK ((x > 0) AND (y < 1)))"]
+  end
+
   specify "should accept named constraint definitions" do
     @db.create_table(:cats) do
       integer :score
@@ -613,13 +665,13 @@ describe "DB#create_table!" do
   end
   
   specify "should create the table if it does not exist" do
-    @db.meta_def(:table_exists?){|a| false}
+    meta_def(@db, :table_exists?){|a| false}
     @db.create_table!(:cats){|*a|}
     @db.sqls.should == ['CREATE TABLE cats ()']
   end
   
   specify "should drop the table before creating it if it already exists" do
-    @db.meta_def(:table_exists?){|a| true}
+    meta_def(@db, :table_exists?){|a| true}
     @db.create_table!(:cats){|*a|}
     @db.sqls.should == ['DROP TABLE cats', 'CREATE TABLE cats ()']
   end
@@ -631,19 +683,19 @@ describe "DB#create_table?" do
   end
   
   specify "should not create the table if the table already exists" do
-    @db.meta_def(:table_exists?){|a| true}
+    meta_def(@db, :table_exists?){|a| true}
     @db.create_table?(:cats){|*a|}
     @db.sqls.should == []
   end
   
   specify "should create the table if the table doesn't already exist" do
-    @db.meta_def(:table_exists?){|a| false}
+    meta_def(@db, :table_exists?){|a| false}
     @db.create_table?(:cats){|*a|}
     @db.sqls.should == ['CREATE TABLE cats ()']
   end
   
   specify "should use IF NOT EXISTS if the database supports that" do
-    @db.meta_def(:supports_create_table_if_not_exists?){true}
+    meta_def(@db, :supports_create_table_if_not_exists?){true}
     @db.create_table?(:cats){|*a|}
     @db.sqls.should == ['CREATE TABLE IF NOT EXISTS cats ()']
   end
@@ -757,37 +809,37 @@ describe "DB#drop_table?" do
   end
   
   specify "should drop the table if it exists" do
-    @db.meta_def(:table_exists?){|a| true}
+    meta_def(@db, :table_exists?){|a| true}
     @db.drop_table?(:cats)
     @db.sqls.should == ["DROP TABLE cats"]
   end
   
   specify "should do nothing if the table does not exist" do
-    @db.meta_def(:table_exists?){|a| false}
+    meta_def(@db, :table_exists?){|a| false}
     @db.drop_table?(:cats)
     @db.sqls.should == []
   end
   
   specify "should operate on multiple tables at once" do
-    @db.meta_def(:table_exists?){|a| a == :cats}
+    meta_def(@db, :table_exists?){|a| a == :cats}
     @db.drop_table? :cats, :dogs
     @db.sqls.should == ['DROP TABLE cats']
   end
 
   specify "should take an options hash and support the :cascade option" do
-    @db.meta_def(:table_exists?){|a| true}
+    meta_def(@db, :table_exists?){|a| true}
     @db.drop_table? :cats, :dogs, :cascade=>true
     @db.sqls.should == ['DROP TABLE cats CASCADE', 'DROP TABLE dogs CASCADE']
   end
 
   specify "should use IF NOT EXISTS if the database supports that" do
-    @db.meta_def(:supports_drop_table_if_exists?){true}
+    meta_def(@db, :supports_drop_table_if_exists?){true}
     @db.drop_table? :cats, :dogs
     @db.sqls.should == ['DROP TABLE IF EXISTS cats', 'DROP TABLE IF EXISTS dogs']
   end
 
   specify "should use IF NOT EXISTS with CASCADE if the database supports that" do
-    @db.meta_def(:supports_drop_table_if_exists?){true}
+    meta_def(@db, :supports_drop_table_if_exists?){true}
     @db.drop_table? :cats, :dogs, :cascade=>true
     @db.sqls.should == ['DROP TABLE IF EXISTS cats CASCADE', 'DROP TABLE IF EXISTS dogs CASCADE']
   end
@@ -896,7 +948,7 @@ describe "DB#alter_table" do
   end
 
   specify "should ignore errors if the database raises an error on an add_index call and the :ignore_errors option is used" do
-    @db.meta_def(:execute_ddl){|*a| raise Sequel::DatabaseError}
+    meta_def(@db, :execute_ddl){|*a| raise Sequel::DatabaseError}
     lambda{@db.add_index(:cats, :id)}.should raise_error(Sequel::DatabaseError)
     lambda{@db.add_index(:cats, :id, :ignore_errors=>true)}.should_not raise_error(Sequel::DatabaseError)
     @db.sqls.should == []
@@ -949,6 +1001,45 @@ describe "DB#alter_table" do
     @db.sqls.should == ["ALTER TABLE cats DROP CONSTRAINT valid_score CASCADE"]
   end
 
+  specify "should support drop_foreign_key" do
+    def @db.foreign_key_list(table_name)
+      [{:name=>:cats_node_id_fkey, :columns=>[:node_id]}] 
+    end
+    @db.alter_table(:cats) do
+      drop_foreign_key :node_id
+    end
+    @db.sqls.should == ["ALTER TABLE cats DROP CONSTRAINT cats_node_id_fkey", "ALTER TABLE cats DROP COLUMN node_id"]
+  end
+
+  specify "should support drop_foreign_key with composite foreign keys" do
+    def @db.foreign_key_list(table_name)
+      [{:name=>:cats_node_id_prop_id_fkey, :columns=>[:node_id, :prop_id]}] 
+    end
+    @db.alter_table(:cats) do
+      drop_foreign_key [:node_id, :prop_id]
+    end
+    @db.sqls.should == ["ALTER TABLE cats DROP CONSTRAINT cats_node_id_prop_id_fkey"]
+
+    @db.alter_table(:cats) do
+      drop_foreign_key [:node_id, :prop_id], :name => :cfk
+    end
+    @db.sqls.should == ["ALTER TABLE cats DROP CONSTRAINT cfk"]
+  end
+
+  specify "should have drop_foreign_key raise Error if no name is found" do
+    def @db.foreign_key_list(table_name)
+      [{:name=>:cats_node_id_fkey, :columns=>[:foo_id]}] 
+    end
+    lambda{@db.alter_table(:cats){drop_foreign_key :node_id}}.should raise_error(Sequel::Error)
+  end
+
+  specify "should have drop_foreign_key raise Error if multiple foreign keys found" do
+    def @db.foreign_key_list(table_name)
+      [{:name=>:cats_node_id_fkey, :columns=>[:node_id]}, {:name=>:cats_node_id_fkey2, :columns=>[:node_id]}] 
+    end
+    lambda{@db.alter_table(:cats){drop_foreign_key :node_id}}.should raise_error(Sequel::Error)
+  end
+
   specify "should support drop_index" do
     @db.alter_table(:cats) do
       drop_index :name
@@ -996,7 +1087,7 @@ describe "DB#alter_table" do
   end
 
   specify "should combine operations into a single query if the database supports it" do
-    @db.meta_def(:supports_combining_alter_table_ops?){true}
+    meta_def(@db, :supports_combining_alter_table_ops?){true}
     @db.alter_table(:cats) do
       add_column :a, Integer
       drop_column :b
@@ -1011,7 +1102,7 @@ describe "DB#alter_table" do
   end
   
   specify "should combine operations into consecutive groups of combinable operations if the database supports combining operations" do
-    @db.meta_def(:supports_combining_alter_table_ops?){true}
+    meta_def(@db, :supports_combining_alter_table_ops?){true}
     @db.alter_table(:cats) do
       add_column :a, Integer
       drop_column :b
@@ -1055,14 +1146,6 @@ describe "Database#create_table" do
     @db.sqls.should == ['CREATE TEMPORARY TABLE test_tmp (id integer NOT NULL PRIMARY KEY AUTOINCREMENT, name text)',
       'CREATE UNIQUE INDEX test_tmp_name_index ON test_tmp (name)']
   end
-
-  specify "should not use default schema when creating a temporary table" do
-    @db.default_schema = :foo
-    @db.create_table :test_tmp, :temp => true do
-      column :name, :text
-    end
-    @db.sqls.should == ['CREATE TEMPORARY TABLE test_tmp (name text)']
-  end
 end
 
 describe "Database#alter_table" do
@@ -1218,15 +1301,30 @@ describe "Database#create_view" do
   specify "should construct proper SQL with dataset" do
     @db.create_view :test, @db[:items].select(:a, :b).order(:c)
     @db.sqls.should == ['CREATE VIEW test AS SELECT a, b FROM items ORDER BY c']
-    @db.create_or_replace_view :sch__test, "SELECT * FROM xyz"
-    @db.sqls.should == ['CREATE OR REPLACE VIEW sch.test AS SELECT * FROM xyz']
   end
 
-  specify "should construct proper SQL with dataset" do
+  specify "should handle :columns option" do
+    @db.create_view :test, @db[:items].select(:a, :b).order(:c), :columns=>[:d, :e]
+    @db.sqls.should == ['CREATE VIEW test (d, e) AS SELECT a, b FROM items ORDER BY c']
+    @db.create_view :test, @db[:items].select(:a, :b).order(:c), :columns=>%w'd e'
+    @db.sqls.should == ['CREATE VIEW test (d, e) AS SELECT a, b FROM items ORDER BY c']
+    @db.create_view :test, @db[:items].select(:a, :b).order(:c), :columns=>[Sequel.identifier('d'), Sequel.lit('e')]
+    @db.sqls.should == ['CREATE VIEW test (d, e) AS SELECT a, b FROM items ORDER BY c']
+  end
+
+  specify "should handle create_or_replace_view" do
+    @db.create_or_replace_view :sch__test, "SELECT * FROM xyz"
+    @db.sqls.should == ['DROP VIEW sch.test', 'CREATE VIEW sch.test AS SELECT * FROM xyz']
     @db.create_or_replace_view :test, @db[:items].select(:a, :b).order(:c)
-    @db.sqls.should == ['CREATE OR REPLACE VIEW test AS SELECT a, b FROM items ORDER BY c']
+    @db.sqls.should == ['DROP VIEW test', 'CREATE VIEW test AS SELECT a, b FROM items ORDER BY c']
     @db.create_or_replace_view Sequel.identifier(:test), @db[:items].select(:a, :b).order(:c)
-    @db.sqls.should == ['CREATE OR REPLACE VIEW test AS SELECT a, b FROM items ORDER BY c']
+    @db.sqls.should == ['DROP VIEW test', 'CREATE VIEW test AS SELECT a, b FROM items ORDER BY c']
+  end
+
+  specify "should use CREATE OR REPLACE VIEW if such syntax is supported" do
+    def @db.supports_create_or_replace_view?() true end
+    @db.create_or_replace_view :test, @db[:items]
+    @db.sqls.should == ['CREATE OR REPLACE VIEW test AS SELECT * FROM items']
   end
 end
 
@@ -1267,21 +1365,21 @@ describe "Schema Parser" do
   end
 
   specify "should raise an error if there are no columns" do
-    @db.meta_def(:schema_parse_table) do |t, opts|
+    meta_def(@db, :schema_parse_table) do |t, opts|
       []
     end
     proc{@db.schema(:x)}.should raise_error(Sequel::Error)
   end
 
   specify "should cache data by default" do
-    @db.meta_def(:schema_parse_table) do |t, opts|
+    meta_def(@db, :schema_parse_table) do |t, opts|
       [[:a, {}]]
     end
     @db.schema(:x).should equal(@db.schema(:x))
   end
 
   specify "should not cache data if :reload=>true is given" do
-    @db.meta_def(:schema_parse_table) do |t, opts|
+    meta_def(@db, :schema_parse_table) do |t, opts|
       [[:a, {}]]
     end
     @db.schema(:x).should_not equal(@db.schema(:x, :reload=>true))
@@ -1289,7 +1387,7 @@ describe "Schema Parser" do
 
   specify "should not cache schema metadata if cache_schema is false" do
     @db.cache_schema = false
-    @db.meta_def(:schema_parse_table) do |t, opts|
+    meta_def(@db, :schema_parse_table) do |t, opts|
       [[:a, {}]]
     end
     @db.schema(:x).should_not equal(@db.schema(:x))
@@ -1297,7 +1395,7 @@ describe "Schema Parser" do
 
   specify "should provide options if given a table name" do
     c = nil
-    @db.meta_def(:schema_parse_table) do |t, opts|
+    meta_def(@db, :schema_parse_table) do |t, opts|
       c = [t, opts]
       [[:a, {:db_type=>t.to_s}]]
     end
@@ -1313,7 +1411,7 @@ describe "Schema Parser" do
   specify "should parse the schema correctly for a single table" do
     sqls = @sqls
     proc{@db.schema(:x)}.should raise_error(Sequel::Error)
-    @db.meta_def(:schema_parse_table) do |t, opts|
+    meta_def(@db, :schema_parse_table) do |t, opts|
       sqls << t
       [[:a, {:db_type=>t.to_s}]]
     end
@@ -1326,7 +1424,7 @@ describe "Schema Parser" do
   end
 
   specify "should convert various types of table name arguments" do
-    @db.meta_def(:schema_parse_table) do |t, opts|
+    meta_def(@db, :schema_parse_table) do |t, opts|
       [[t, opts]]
     end
     s1 = @db.schema(:x)
@@ -1387,7 +1485,6 @@ describe "Schema Parser" do
     @db.schema(:"number(10,0)").first.last[:type].should == :integer
     @db.schema(:"numeric(10, 10)").first.last[:type].should == :decimal
     @db.schema(:"decimal(10,1)").first.last[:type].should == :decimal
-    @db.schema(:money).first.last[:type].should == :decimal
     @db.schema(:bytea).first.last[:type].should == :blob
     @db.schema(:blob).first.last[:type].should == :blob
     @db.schema(:image).first.last[:type].should == :blob
@@ -1395,7 +1492,6 @@ describe "Schema Parser" do
     @db.schema(:nvarchar).first.last[:type].should == :string
     @db.schema(:ntext).first.last[:type].should == :string
     @db.schema(:smalldatetime).first.last[:type].should == :datetime
-    @db.schema(:smallmoney).first.last[:type].should == :decimal
     @db.schema(:binary).first.last[:type].should == :blob
     @db.schema(:varbinary).first.last[:type].should == :blob
     @db.schema(:enum).first.last[:type].should == :enum
@@ -1407,5 +1503,7 @@ describe "Schema Parser" do
     @db = Sequel.mock(:host=>'mysql')
     @db.extend(sm)
     @db.schema(:set).first.last[:type].should == :set
+    @db.schema(:mediumint).first.last[:type].should == :integer
+    @db.schema(:mediumtext).first.last[:type].should == :string
   end
 end
