@@ -1,5 +1,7 @@
 require File.join(File.dirname(File.expand_path(__FILE__)), 'spec_helper')
 
+Sequel.extension :migration
+
 describe "Migration.descendants" do
   before do
     Sequel::Migration.descendants.clear
@@ -55,6 +57,12 @@ describe "Migration.apply" do
     m.apply(@db, :up).should == nil
     m.apply(@db, :down).should == nil
   end
+
+  specify "should respond to the methods the database responds to" do
+    m = Sequel::Migration.new(Sequel.mock)
+    m.respond_to?(:foo).should be_false
+    m.respond_to?(:execute).should be_true
+  end if RUBY_VERSION >= '1.9'
 end
 
 describe "SimpleMigration#apply" do
@@ -122,6 +130,8 @@ describe "Reversible Migrations with Sequel.migration{change{}}" do
         add_column :d, String
         add_constraint :blah, 'd IS NOT NULL'
         add_foreign_key :e, :b
+        add_foreign_key [:e], :b, :name=>'e_fk'
+        add_foreign_key [:e, :a], :b
         add_primary_key :f, :b
         add_index :e, :name=>'e_n'
         add_full_text_index :e, :name=>'e_ft'
@@ -129,6 +139,7 @@ describe "Reversible Migrations with Sequel.migration{change{}}" do
         rename_column :e, :g
       end
       create_view(:c, 'SELECT * FROM b')
+      create_join_table(:cat_id=>:cats, :dog_id=>:dogs)
     end
   end
   
@@ -144,26 +155,33 @@ describe "Reversible Migrations with Sequel.migration{change{}}" do
         [:add_column, :d, String],
         [:add_constraint, :blah, "d IS NOT NULL"],
         [:add_foreign_key, :e, :b],
+        [:add_foreign_key, [:e], :b, {:name=>"e_fk"}],
+        [:add_foreign_key, [:e, :a], :b],
         [:add_primary_key, :f, :b],
         [:add_index, :e, {:name=>"e_n"}],
         [:add_full_text_index, :e, {:name=>"e_ft"}],
         [:add_spatial_index, :e, {:name=>"e_s"}],
         [:rename_column, :e, :g]]
       ],
-      [:create_view, :c, "SELECT * FROM b"]]
+      [:create_view, :c, "SELECT * FROM b"],
+      [:create_join_table, {:cat_id=>:cats, :dog_id=>:dogs}]]
   end
 
   specify "should execute down with reversing actions in reverse order" do
     p = @p
     Sequel.migration{change(&p)}.apply(@db, :down)
-    @db.actions.should == [[:drop_view, :c],
+    @db.actions.should == [
+      [:drop_join_table, {:cat_id=>:cats, :dog_id=>:dogs}],
+      [:drop_view, :c],
       [:alter_table, [
         [:rename_column, :g, :e],
         [:drop_index, :e, {:name=>"e_s"}],
         [:drop_index, :e, {:name=>"e_ft"}],
         [:drop_index, :e, {:name=>"e_n"}],
         [:drop_column, :f],
-        [:drop_column, :e],
+        [:drop_foreign_key, [:e, :a]],
+        [:drop_foreign_key, [:e], {:name=>"e_fk"}],
+        [:drop_foreign_key, :e],
         [:drop_constraint, :blah],
         [:drop_column, :d]]
       ],
@@ -195,7 +213,7 @@ end
 
 describe "Sequel::IntegerMigrator" do
   before do
-    dbc = Class.new(MockDatabase) do
+    dbc = Class.new(Sequel::Mock::Database) do
       attr_reader :drops, :tables_created, :columns_created, :versions
       def initialize(*args)
         super
@@ -211,17 +229,17 @@ describe "Sequel::IntegerMigrator" do
 
       def create_table(name, opts={}, &block)
         super
-        @columns_created << / \(?(\w+) integer.*\)?\z/.match(sqls.last)[1].to_sym
+        @columns_created << / \(?(\w+) integer.*\)?\z/.match(@sqls.last)[1].to_sym
         @tables_created << name.to_sym
       end
       
-      def dataset(opts={})
+      def dataset
         ds = super
         ds.extend(Module.new do
           def count; 1; end
           def columns; db.columns_created end
-          def insert(h); db.versions.merge!(h); db.sqls << insert_sql(h) end
-          def update(h); db.versions.merge!(h); db.sqls << update_sql(h) end
+          def insert(h); db.versions.merge!(h); db.run insert_sql(h) end
+          def update(h); db.versions.merge!(h); db.run update_sql(h) end
           def fetch_rows(sql); db.execute(sql); yield(db.versions) unless db.versions.empty? end
         end)
         ds
@@ -243,6 +261,10 @@ describe "Sequel::IntegerMigrator" do
   specify "should raise and error if there is a missing integer migration version" do
     proc{Sequel::Migrator.apply(@db, "spec/files/missing_integer_migrations")}.should raise_error(Sequel::Migrator::Error)
   end
+  
+  specify "should not raise and error if there is a missing integer migration version and allow_missing_migration_files is true" do
+    Sequel::Migrator.run(@db, "spec/files/missing_integer_migrations", :allow_missing_migration_files => true).should_not raise_error(Sequel::Migrator::Error)
+  end
 
   specify "should raise and error if there is a duplicate integer migration version" do
     proc{Sequel::Migrator.apply(@db, "spec/files/duplicate_integer_migrations")}.should raise_error(Sequel::Migrator::Error)
@@ -273,8 +295,20 @@ describe "Sequel::IntegerMigrator" do
     @db.creates.should == [1111, 2222, 3333]
     @db.version.should == 3
     @db.sqls.map{|x| x =~ /\AUPDATE.*(\d+)/ ? $1.to_i : nil}.compact.should == [1, 2, 3]
+  end
+  
+  specify "should be able to tell whether there are outstanding migrations" do
+    Sequel::Migrator.is_current?(@db, @dirname).should be_false
+    Sequel::Migrator.apply(@db, @dirname)
+    Sequel::Migrator.is_current?(@db, @dirname).should be_true
   end 
 
+  specify "should have #check_current raise an exception if the migrator is not current" do
+    proc{Sequel::Migrator.check_current(@db, @dirname)}.should raise_error(Sequel::Migrator::NotCurrentError)
+    Sequel::Migrator.apply(@db, @dirname)
+    proc{Sequel::Migrator.check_current(@db, @dirname)}.should_not raise_error
+  end
+
   specify "should apply migrations correctly in the up direction with target" do
     Sequel::Migrator.apply(@db, @dirname, 2)
     @db.creates.should == [1111, 2222]
@@ -311,14 +345,49 @@ describe "Sequel::IntegerMigrator" do
     Sequel::Migrator.apply(@db, @dirname, 0).should == 0
     Sequel::Migrator.apply(@db, @dirname).should == 3
   end
+
+  specify "should use IntegerMigrator if IntegerMigrator.apply called, even for timestamped migration directory" do
+    proc{Sequel::IntegerMigrator.apply(@db, "spec/files/timestamped_migrations")}.should raise_error(Sequel::Migrator::Error)
+  end
+
+  specify "should not use transactions by default" do
+    Sequel::Migrator.apply(@db, "spec/files/transaction_unspecified_migrations")
+    @db.sqls.should == ["CREATE TABLE schema_info (version integer DEFAULT 0 NOT NULL)", "SELECT 1 AS one FROM schema_info LIMIT 1", "INSERT INTO schema_info (version) VALUES (0)", "SELECT version FROM schema_info LIMIT 1", "CREATE TABLE sm11111 (smc1 integer)", "UPDATE schema_info SET version = 1", "CREATE TABLE sm (smc1 integer)", "UPDATE schema_info SET version = 2"]
+  end
+
+  specify "should use transactions by default if the database supports transactional ddl" do
+    @db.meta_def(:supports_transactional_ddl?){true}
+    Sequel::Migrator.apply(@db, "spec/files/transaction_unspecified_migrations")
+    @db.sqls.should == ["CREATE TABLE schema_info (version integer DEFAULT 0 NOT NULL)", "SELECT 1 AS one FROM schema_info LIMIT 1", "INSERT INTO schema_info (version) VALUES (0)", "SELECT version FROM schema_info LIMIT 1", "BEGIN", "CREATE TABLE sm11111 (smc1 integer)", "UPDATE schema_info SET version = 1", "COMMIT", "BEGIN", "CREATE TABLE sm (smc1 integer)", "UPDATE schema_info SET version = 2", "COMMIT"]
+  end
+
+  specify "should respect transaction use on a per migration basis" do
+    @db.meta_def(:supports_transactional_ddl?){true}
+    Sequel::Migrator.apply(@db, "spec/files/transaction_specified_migrations")
+    @db.sqls.should == ["CREATE TABLE schema_info (version integer DEFAULT 0 NOT NULL)", "SELECT 1 AS one FROM schema_info LIMIT 1", "INSERT INTO schema_info (version) VALUES (0)", "SELECT version FROM schema_info LIMIT 1", "BEGIN", "CREATE TABLE sm11111 (smc1 integer)", "UPDATE schema_info SET version = 1", "COMMIT", "CREATE TABLE sm (smc1 integer)", "UPDATE schema_info SET version = 2"]
+  end
+
+  specify "should force transactions if enabled in the migrator" do
+    Sequel::Migrator.run(@db, "spec/files/transaction_specified_migrations", :use_transactions=>true)
+    @db.sqls.should == ["CREATE TABLE schema_info (version integer DEFAULT 0 NOT NULL)", "SELECT 1 AS one FROM schema_info LIMIT 1", "INSERT INTO schema_info (version) VALUES (0)", "SELECT version FROM schema_info LIMIT 1", "BEGIN", "CREATE TABLE sm11111 (smc1 integer)", "UPDATE schema_info SET version = 1", "COMMIT", "BEGIN", "CREATE TABLE sm (smc1 integer)", "UPDATE schema_info SET version = 2", "COMMIT"]
+  end
+
+  specify "should not use transactions if disabled in the migrator" do
+    Sequel::Migrator.run(@db, "spec/files/transaction_unspecified_migrations", :use_transactions=>false)
+    @db.sqls.should == ["CREATE TABLE schema_info (version integer DEFAULT 0 NOT NULL)", "SELECT 1 AS one FROM schema_info LIMIT 1", "INSERT INTO schema_info (version) VALUES (0)", "SELECT version FROM schema_info LIMIT 1", "CREATE TABLE sm11111 (smc1 integer)", "UPDATE schema_info SET version = 1", "CREATE TABLE sm (smc1 integer)", "UPDATE schema_info SET version = 2"]
+  end
 end
 
 describe "Sequel::TimestampMigrator" do
   before do
-    $sequel_migration_version = 0
-    $sequel_migration_files = []
-    @dsc = dsc = Class.new(MockDataset) do
+    sequel_migration_version = 0
+    @dsc = dsc = Class.new(Sequel::Mock::Dataset) do
+      self::FILES =[]
+      define_method(:sequel_migration_version){sequel_migration_version}
+      define_method(:sequel_migration_version=){|v| sequel_migration_version = v}
+
       def columns
+        super
         case opts[:from].first
         when :schema_info, 'schema_info'
           [:version]
@@ -330,45 +399,52 @@ describe "Sequel::TimestampMigrator" do
       end
 
       def fetch_rows(sql)
+        super
         case opts[:from].first
         when :schema_info, 'schema_info'
-          yield({:version=>$sequel_migration_version})
+          yield({:version=>sequel_migration_version})
         when :schema_migrations, 'schema_migrations'
-          $sequel_migration_files.sort.each{|f| yield(:filename=>f)}
+          self.class::FILES.sort.each{|f| yield(:filename=>f)}
         when :sm, 'sm'
-          $sequel_migration_files.sort.each{|f| yield(:fn=>f)}
+          self.class::FILES.sort.each{|f| yield(:fn=>f)}
         end
       end
 
       def insert(h={})
+        super
         case opts[:from].first
         when :schema_info, 'schema_info'
-          $sequel_migration_version = h.values.first
+          self.sequel_migration_version = h.values.first
         when :schema_migrations, :sm, 'schema_migrations', 'sm'
-          $sequel_migration_files << h.values.first
+          self.class::FILES << h.values.first
         end
       end
 
       def update(h={})
+        super
         case opts[:from].first
         when :schema_info, 'schema_info'
-          $sequel_migration_version = h.values.first
+          self.sequel_migration_version = h.values.first
         end
       end
 
       def delete
+        super
         case opts[:from].first
         when :schema_migrations, :sm, 'schema_migrations', 'sm'
-          $sequel_migration_files.delete(opts[:where].args.last)
+          self.class::FILES.delete(opts[:where].args.last)
         end
       end
     end
-    dbc = Class.new(MockDatabase) do
-      tables = {}
+    dbc = Class.new(Sequel::Mock::Database) do
+      self::Tables = tables= {}
       define_method(:dataset){|*a| dsc.new(self, *a)}
-      define_method(:create_table){|name, *args| tables[name.to_sym] = true}
-      define_method(:drop_table){|*names| names.each{|n| tables.delete(n.to_sym)}}
-      define_method(:table_exists?){|name| tables.has_key?(name.to_sym)}
+      def create_table(name, *args, &block)
+        super
+        self.class::Tables[name.to_sym] = true
+      end
+      define_method(:drop_table){|*names| super(*names); names.each{|n| tables.delete(n.to_sym)}}
+      define_method(:table_exists?){|name| super(name); tables.has_key?(name.to_sym)}
     end
     @db = dbc.new
     @m = Sequel::Migrator
@@ -402,6 +478,22 @@ describe "Sequel::TimestampMigrator" do
     @db[:schema_migrations].select_order_map(:filename).should == %w'1273253849_create_sessions.rb'
   end
 
+  specify "should not be current when there are migrations to apply" do
+    @dir = 'spec/files/timestamped_migrations'
+    @m.apply(@db, @dir)
+    @m.is_current?(@db, @dir).should be_true
+    @dir = 'spec/files/interleaved_timestamped_migrations'
+    @m.is_current?(@db, @dir).should be_false
+  end
+
+  specify "should raise an exception if the migrator is not current" do
+    @dir = 'spec/files/timestamped_migrations'
+    @m.apply(@db, @dir)
+    proc{@m.check_current(@db, @dir)}.should_not raise_error
+    @dir = 'spec/files/interleaved_timestamped_migrations'
+    proc{@m.check_current(@db, @dir)}.should raise_error(Sequel::Migrator::NotCurrentError)
+  end
+
   specify "should apply all missing files when migrating up" do
     @dir = 'spec/files/timestamped_migrations'
     @m.apply(@db, @dir)
@@ -523,6 +615,18 @@ describe "Sequel::TimestampMigrator" do
     @db[:schema_migrations].select_order_map(:filename).should == %w'1273253849_create_sessions.rb 1273253851_create_nodes.rb 1273253853_3_create_users.rb'
   end
   
+  specify "should not raise error for applied migrations not in file system if :allow_missing_migration_files is true" do
+    @dir = 'spec/files/timestamped_migrations'
+    @m.apply(@db, @dir)
+    [:schema_migrations, :sm1111, :sm2222, :sm3333].each{|n| @db.table_exists?(n).should be_true}
+    @db[:schema_migrations].select_order_map(:filename).should == %w'1273253849_create_sessions.rb 1273253851_create_nodes.rb 1273253853_3_create_users.rb'
+
+    @dir = 'spec/files/missing_timestamped_migrations'
+    proc{@m.run(@db, @dir, :allow_missing_migration_files => true)}.should_not raise_error(Sequel::Migrator::Error)
+    [:schema_migrations, :sm1111, :sm2222, :sm3333].each{|n| @db.table_exists?(n).should be_true}
+    @db[:schema_migrations].select_order_map(:filename).should == %w'1273253849_create_sessions.rb 1273253851_create_nodes.rb 1273253853_3_create_users.rb'
+  end
+  
   specify "should raise error missing column name in existing schema_migrations table" do
     @dir = 'spec/files/timestamped_migrations'
     @m.apply(@db, @dir)
@@ -556,4 +660,35 @@ describe "Sequel::TimestampMigrator" do
     @m.apply(@db, @dir, 0).should == nil
     @m.apply(@db, @dir).should == nil
   end
+
+  specify "should use TimestampMigrator if TimestampMigrator.apply is called even for integer migrations directory" do
+    Sequel::TimestampMigrator.apply(@db, "spec/files/integer_migrations")
+    @db.sqls.should == ["SELECT NULL AS nil FROM schema_migrations LIMIT 1", "CREATE TABLE schema_migrations (filename varchar(255) PRIMARY KEY)", "SELECT NULL AS nil FROM schema_info LIMIT 1", "SELECT filename FROM schema_migrations ORDER BY filename", "CREATE TABLE sm1111 (smc1 integer)", "INSERT INTO schema_migrations (filename) VALUES ('001_create_sessions.rb')", "CREATE TABLE sm2222 (smc2 integer)", "INSERT INTO schema_migrations (filename) VALUES ('002_create_nodes.rb')", "CREATE TABLE sm3333 (smc3 integer)", "INSERT INTO schema_migrations (filename) VALUES ('003_3_create_users.rb')"]
+  end
+
+  specify "should not use transactions by default" do
+    Sequel::TimestampMigrator.apply(@db, "spec/files/transaction_unspecified_migrations")
+    @db.sqls.should == ["SELECT NULL AS nil FROM schema_migrations LIMIT 1", "CREATE TABLE schema_migrations (filename varchar(255) PRIMARY KEY)", "SELECT NULL AS nil FROM schema_info LIMIT 1", "SELECT filename FROM schema_migrations ORDER BY filename", "CREATE TABLE sm11111 (smc1 integer)", "INSERT INTO schema_migrations (filename) VALUES ('001_create_alt_basic.rb')", "CREATE TABLE sm (smc1 integer)", "INSERT INTO schema_migrations (filename) VALUES ('002_create_basic.rb')"]
+  end
+
+  specify "should use transactions by default if database supports transactional ddl" do
+    @db.meta_def(:supports_transactional_ddl?){true}
+    Sequel::TimestampMigrator.apply(@db, "spec/files/transaction_unspecified_migrations")
+    @db.sqls.should == ["SELECT NULL AS nil FROM schema_migrations LIMIT 1", "CREATE TABLE schema_migrations (filename varchar(255) PRIMARY KEY)", "SELECT NULL AS nil FROM schema_info LIMIT 1", "SELECT filename FROM schema_migrations ORDER BY filename", "BEGIN", "CREATE TABLE sm11111 (smc1 integer)", "INSERT INTO schema_migrations (filename) VALUES ('001_create_alt_basic.rb')", "COMMIT", "BEGIN", "CREATE TABLE sm (smc1 integer)", "INSERT INTO schema_migrations (filename) VALUES ('002_create_basic.rb')", "COMMIT"]
+  end
+
+  specify "should support transaction use on a per migration basis" do
+    Sequel::TimestampMigrator.apply(@db, "spec/files/transaction_specified_migrations")
+    @db.sqls.should == ["SELECT NULL AS nil FROM schema_migrations LIMIT 1", "CREATE TABLE schema_migrations (filename varchar(255) PRIMARY KEY)", "SELECT NULL AS nil FROM schema_info LIMIT 1", "SELECT filename FROM schema_migrations ORDER BY filename", "BEGIN", "CREATE TABLE sm11111 (smc1 integer)", "INSERT INTO schema_migrations (filename) VALUES ('001_create_alt_basic.rb')", "COMMIT", "CREATE TABLE sm (smc1 integer)", "INSERT INTO schema_migrations (filename) VALUES ('002_create_basic.rb')"]
+  end
+
+  specify "should force transactions if enabled by the migrator" do
+    Sequel::TimestampMigrator.run(@db, "spec/files/transaction_specified_migrations", :use_transactions=>true)
+    @db.sqls.should == ["SELECT NULL AS nil FROM schema_migrations LIMIT 1", "CREATE TABLE schema_migrations (filename varchar(255) PRIMARY KEY)", "SELECT NULL AS nil FROM schema_info LIMIT 1", "SELECT filename FROM schema_migrations ORDER BY filename", "BEGIN", "CREATE TABLE sm11111 (smc1 integer)", "INSERT INTO schema_migrations (filename) VALUES ('001_create_alt_basic.rb')", "COMMIT", "BEGIN", "CREATE TABLE sm (smc1 integer)", "INSERT INTO schema_migrations (filename) VALUES ('002_create_basic.rb')", "COMMIT"]
+  end
+
+  specify "should not use transactions if disabled in the migrator" do
+    Sequel::TimestampMigrator.run(@db, "spec/files/transaction_unspecified_migrations", :use_transactions=>false)
+    @db.sqls.should == ["SELECT NULL AS nil FROM schema_migrations LIMIT 1", "CREATE TABLE schema_migrations (filename varchar(255) PRIMARY KEY)", "SELECT NULL AS nil FROM schema_info LIMIT 1", "SELECT filename FROM schema_migrations ORDER BY filename", "CREATE TABLE sm11111 (smc1 integer)", "INSERT INTO schema_migrations (filename) VALUES ('001_create_alt_basic.rb')", "CREATE TABLE sm (smc1 integer)", "INSERT INTO schema_migrations (filename) VALUES ('002_create_basic.rb')"]
+  end
 end
